-- Practical typed lazy contracts
-- Implementation without informative error messages and without Template Haskell
-- demonstrates how contracts work
-- see: Olaf Chitil "Practical Typed Lazy Contracts", ICFP 2012, ACM Press.

module Development.ContractSimple where

import Data.Maybe(maybe)
import Control.Monad(mplus)

type Contract a = a -> Maybe a

assert :: Contract a -> (a -> a)
assert c x = case c x of
               Just x  -> x
               Nothing -> error "contract failed"

attach :: a -> Contract a -> a
attach x c = assert c x


(&) :: Contract a -> Contract a -> Contract a
c1 & c2 = \x -> c1 x >>= c2

(|>) :: Contract a -> Contract a -> Contract a
c1 |> c2 = \x -> c1 x `mplus` c2 x

(>->) :: Contract a -> Contract b -> Contract (a -> b)
pre >-> post = \f -> Just (f `seq` (assert post . f . assert pre))

-- The dependent function contract combinator does not protect laziness!
(>>->) :: Contract a -> (a -> Contract b) -> Contract (a -> b)
pre >>-> post = 
  \f -> Just (f `seq` \x -> let y = assert pre x in assert (post y) (f y))


true :: Contract a
true = Just

false :: Contract a
false = const Nothing


class Flat a where
  prop :: (a -> Bool) -> Contract a
  prop p = \x -> if p x then Just x else Nothing

instance Flat Int
instance Flat Integer
instance Flat Float
instance Flat Double
instance Flat Char
instance Flat Bool

pNil :: Contract [a]
pNil [] = Just []
pNil (_:_) = Nothing

pCons :: Contract a -> Contract [a] -> Contract [a]
pCons cx cxs (z:zs) = Just (assert cx z : assert cxs zs)
pCons cx cxs [] = Nothing

pNotNil :: Contract [a]
pNotNil [] = Nothing
pNotNil xs = Just xs

pNotCons :: Contract [a]
pNotCons (_:_) = Nothing
pNotCons xs = Just xs

list :: Contract a -> Contract [a]
list c = pNil |> pCons c (list c)

io :: Contract a -> Contract (IO a)
io c m = m `seq` Just (m >>= return . assert c)

p0 :: Contract a
p0 = Just

p1 :: (a -> b) -> Contract a -> a -> Maybe b
p1 c c1 x1 = Just (c (assert c1 x1))

p2 :: (a -> b -> c) -> Contract a -> Contract b -> a -> b -> Maybe c
p2 c c1 c2 x1 x2 = Just (c (assert c1 x1) (assert c2 x2))

p3 :: (a -> b -> c -> d) -> Contract a -> Contract b -> Contract c -> 
      a -> b -> c -> Maybe d
p3 c c1 c2 c3 x1 x2 x3 = Just (c (assert c1 x1) (assert c2 x2) (assert c3 x3))

pTuple0 :: Contract ()
pTuple0 () = p0  ()

pTuple2 :: Contract a -> Contract b -> Contract (a,b)
pTuple2 c1 c2 (x1,x2) = p2 (,) c1 c2 x1 x2

pTuple3 :: Contract a -> Contract b -> Contract c -> Contract (a,b,c)
pTuple3 c1 c2 c3 (x1,x2,x3) = p3 (,,) c1 c2 c3 x1 x2 x3

pNothing :: Contract (Maybe a)
pNothing Nothing = p0 Nothing
pNothing _ = Nothing

pJust :: Contract a -> Contract (Maybe a)
pJust c (Just x) = p1 Just c x
pJust c _ = Nothing

pNotNothing :: Contract (Maybe a)
pNotNothing Nothing = Nothing
pNotNothing x = Just x

pNotJust :: Contract (Maybe a)
pNotJust (Just _) = Nothing
pNotJust x = Just x

maybe :: Contract a -> Contract (Maybe a)
maybe c = pNothing |> pJust c 
