-- Example use of contracts

{-# OPTIONS_GHC -XDatatypeContexts #-}
-- ghc needs option to allow H98 class context in a data type

module Development.ContractSimpleBlameTest where

import Development.ContractSimpleBlame

nat :: (Integral a, Flat a) => Contract a
nat = prop (>= 0)

nats :: Contract [Integer]
nats = list nat

fibs :: [Integer]
fibs = assert nats
         (0 : 1 : zipWith (+) fibs (tail fibs))

infinite :: Contract [a]
infinite = pCons true infinite

lengthAtLeast :: Int -> Contract [a]
lengthAtLeast 0 = true
lengthAtLeast n = pCons true (lengthAtLeast (n-1))

initAv :: [Int] -> Int
initAv = assert (lengthAtLeast 5 >-> true) initAv'

initAv' xs = sum (take 5 xs) `div` 5

listOfLength :: Int -> Contract [a]
listOfLength 0 = pNil
listOfLength n = pCons true (listOfLength (n-1))

contractTake :: Int -> [a] -> [a]
contractTake = 
  assert (nat >>-> \n -> lengthAtLeast n >-> listOfLength n) take


data Form = Imp Form Form | And Form Form |
            Or Form Form | Not Form | Atom Char
  deriving Show

pImp :: Contract Form -> Contract Form -> Contract Form
pImp c1 c2 = \f -> case f of
                     Imp f1 f2 -> p2 Imp c1 c2 f1 f2
                     _         -> cFail

pAnd :: Contract Form -> Contract Form -> Contract Form
pAnd c1 c2 = \f -> case f of
                     And f1 f2 -> p2 And c1 c2 f1 f2
                     _         -> cFail

pOr :: Contract Form -> Contract Form -> Contract Form
pOr c1 c2 = \f -> case f of
                    Or f1 f2 -> p2 Or c1 c2 f1 f2
                    _        -> cFail

pNot :: Contract Form -> Contract Form
pNot c1 = \f -> case f of
                  Not f1 -> p1 Not c1 f1
                  _      -> cFail

pAtom :: Contract Char -> Contract Form
pAtom c1 = \f -> case f of
                   Atom a -> p1 Atom c1 a
                   _      -> cFail

pNotImp :: Contract Form
pNotImp = \f -> case f of
                  Imp _ _ -> cFail
                  _       -> p0 f

pNotAnd :: Contract Form
pNotAnd = \f -> case f of
                  And _ _ -> cFail
                  _       -> p0 f

pNotOr :: Contract Form
pNotOr = \f -> case f of
                 Or _ _ -> cFail
                 _      -> p0 f

pNotNot :: Contract Form
pNotNot = \f -> case f of
                  Not _ -> cFail
                  _     -> p0 f

pNotAtom :: Contract Form
pNotAtom = \f -> case f of
                   Atom _ -> cFail
                   _      -> p0 f



conjNF, disj, lit, atom :: Contract Form

conjNF = pAnd conjNF conjNF |> disj
disj = pOr disj disj |> lit
lit = pNot atom |> atom
atom = pAtom true

right = pImp (right & pNotImp) right |>
       pAnd (right & pNotAnd) right |>
       pOr (right & pNotOr) right |>
       pNot right |> pAtom true

rightConjNF :: Contract Form
rightConjNF = conjNF & right

-- tests:

f1 = Not (Atom 'a')
f2 = Not f1
f3 = And (Or (Atom 'a') f1) (Or f1 (Atom 'b'))
f4 = Or f1 f3
f5 = Or (Atom 'a') (And (Atom 'b') (Atom 'c'))
f6 = And (Atom 'a') (And (Atom 'b') (Atom 'c')) 

c1 = assert conjNF f1
c2 = assert conjNF f2  -- fail
c3 = assert conjNF f3
c4 = assert conjNF f4  -- fail
c5 = assert conjNF f5  -- fail

l1 = assert rightConjNF f1
l2 = assert rightConjNF f2  -- fail
l3 = assert rightConjNF f3
l4 = assert rightConjNF f4  -- fail
l5 = assert rightConjNF f5  -- fail
l6 = assert rightConjNF f6


clausalNF = attach clausalNF' (rightConjNF >-> list (list lit))

clausalNF' :: Form -> [[Form]]
clausalNF' (And f1 f2) = clause f1 : clausalNF' f2
clausalNF' f = [clause f]

clause = attach clause' ((disj & right) >-> list lit)

clause' :: Form -> [Form]
clause' (Or f1 f2) = f1 : clause' f2
clause' lit = [lit] 


-- tree example

data Ord a => Tree a = Leaf a | Tree a :^: Tree a
  deriving Show 

pLeaf :: Ord a => Contract a -> Contract (Tree a)
pLeaf cx = \x -> case x of
                   Leaf y  -> p1 Leaf cx y
                   (_:^:_) -> cFail

(=:^:) :: Ord a => Contract (Tree a) -> Contract (Tree a) -> Contract (Tree a)
cx =:^: cy = \z -> case z of
                     (x:^:y) -> p2 (:^:) cx cy x y
                     Leaf _  -> cFail

pNotLeaf :: Ord a => Contract (Tree a)
pNotLeaf = \x -> case x of
                   Leaf _ -> cFail
                   _      -> p0 x


o1 = assert (pLeaf true =:^: pLeaf true) (Leaf 4 :^: (Leaf 5 :^: Leaf 7))

o2 = assert pNotLeaf (Leaf 7)
